@@ -36,7 +36,7 @@ try:
 except NameError:  # Python 3
     basestring = str
 
-_probe_limit = 1000
+_default_probe_limit = 1000
 _num_open_probes = 0
 
 # for tests
@@ -104,7 +104,6 @@ class SymbolCache(object):
             return -1
         return addr.value
 
-
 class PerfType:
     # From perf_type_id in uapi/linux/perf_event.h
     HARDWARE = 0
@@ -189,6 +188,7 @@ class BPFProgType:
     SK_MSG = 16
     RAW_TRACEPOINT = 17
     CGROUP_SOCK_ADDR = 18
+    CGROUP_SOCKOPT = 25
     TRACING = 26
     LSM = 29
 
@@ -450,32 +450,28 @@ class BPF(object):
             src_file = BPF._find_file(src_file)
             hdr_file = BPF._find_file(hdr_file)
 
-        # files that end in ".b" are treated as B files. Everything else is a (BPF-)C file
-        if src_file.endswith(b".b"):
-            self.module = lib.bpf_module_create_b(src_file, hdr_file, self.debug, device)
-        else:
-            if src_file:
-                # Read the BPF C source file into the text variable. This ensures,
-                # that files and inline text are treated equally.
-                with open(src_file, mode="rb") as file:
-                    text = file.read()
+        if src_file:
+            # Read the BPF C source file into the text variable. This ensures,
+            # that files and inline text are treated equally.
+            with open(src_file, mode="rb") as file:
+                text = file.read()
 
-            ctx_array = (ct.c_void_p * len(usdt_contexts))()
-            for i, usdt in enumerate(usdt_contexts):
-                ctx_array[i] = ct.c_void_p(usdt.get_context())
-            usdt_text = lib.bcc_usdt_genargs(ctx_array, len(usdt_contexts))
-            if usdt_text is None:
-                raise Exception("can't generate USDT probe arguments; " +
-                                "possible cause is missing pid when a " +
-                                "probe in a shared object has multiple " +
-                                "locations")
-            text = usdt_text + text
+        ctx_array = (ct.c_void_p * len(usdt_contexts))()
+        for i, usdt in enumerate(usdt_contexts):
+            ctx_array[i] = ct.c_void_p(usdt.get_context())
+        usdt_text = lib.bcc_usdt_genargs(ctx_array, len(usdt_contexts))
+        if usdt_text is None:
+            raise Exception("can't generate USDT probe arguments; " +
+                            "possible cause is missing pid when a " +
+                            "probe in a shared object has multiple " +
+                            "locations")
+        text = usdt_text + text
 
 
-            self.module = lib.bpf_module_create_c_from_string(text,
-                                                              self.debug,
-                                                              cflags_array, len(cflags_array),
-                                                              allow_rlimit, device)
+        self.module = lib.bpf_module_create_c_from_string(text,
+                                                          self.debug,
+                                                          cflags_array, len(cflags_array),
+                                                          allow_rlimit, device)
         if not self.module:
             raise Exception("Failed to compile BPF module %s" % (src_file or "<text>"))
 
@@ -500,7 +496,7 @@ class BPF(object):
 
         return fns
 
-    def load_func(self, func_name, prog_type, device = None):
+    def load_func(self, func_name, prog_type, device = None, attach_type = -1):
         func_name = _assert_is_bytes(func_name)
         if func_name in self.funcs:
             return self.funcs[func_name]
@@ -516,7 +512,7 @@ class BPF(object):
                 lib.bpf_function_size(self.module, func_name),
                 lib.bpf_module_license(self.module),
                 lib.bpf_module_kern_version(self.module),
-                log_level, None, 0, device)
+                log_level, None, 0, device, attach_type)
 
         if fd < 0:
             atexit.register(self.donothing)
@@ -742,6 +738,10 @@ class BPF(object):
                 # non-attachable.
                 elif fn.startswith(b'__perf') or fn.startswith(b'perf_'):
                     continue
+                # Exclude all static functions with prefix __SCT__, they are
+                # all non-attachable
+                elif fn.startswith(b'__SCT__'):
+                    continue
                 # Exclude all gcc 8's extra .cold functions
                 elif re.match(b'^.*\.cold(\.\d+)?$', fn):
                     continue
@@ -752,9 +752,17 @@ class BPF(object):
 
     def _check_probe_quota(self, num_new_probes):
         global _num_open_probes
-        if _num_open_probes + num_new_probes > _probe_limit:
+        if _num_open_probes + num_new_probes > BPF.get_probe_limit():
             raise Exception("Number of open probes would exceed global quota")
 
+    @staticmethod
+    def get_probe_limit():
+        env_probe_limit = os.environ.get('BCC_PROBE_LIMIT')
+        if env_probe_limit and env_probe_limit.isdigit():
+            return int(env_probe_limit)
+        else:
+            return _default_probe_limit
+
     def _add_kprobe_fd(self, ev_name, fn_name, fd):
         global _num_open_probes
         if ev_name not in self.kprobe_fds:
@@ -821,7 +829,8 @@ class BPF(object):
                     failed += 1
                     probes.append(line)
             if failed == len(matches):
-                raise Exception("Failed to attach BPF program %s to kprobe %s" %
+                raise Exception("Failed to attach BPF program %s to kprobe %s"
+                                ", it's not traceable (either non-existing, inlined, or marked as \"notrace\")" %
                                 (fn_name, '/'.join(probes)))
             return
 
@@ -830,7 +839,8 @@ class BPF(object):
         ev_name = b"p_" + event.replace(b"+", b"_").replace(b".", b"_")
         fd = lib.bpf_attach_kprobe(fn.fd, 0, ev_name, event, event_off, 0)
         if fd < 0:
-            raise Exception("Failed to attach BPF program %s to kprobe %s" %
+            raise Exception("Failed to attach BPF program %s to kprobe %s"
+                            ", it's not traceable (either non-existing, inlined, or marked as \"notrace\")" %
                             (fn_name, event))
         self._add_kprobe_fd(ev_name, fn_name, fd)
         return self
@@ -853,7 +863,8 @@ class BPF(object):
                     failed += 1
                     probes.append(line)
             if failed == len(matches):
-                raise Exception("Failed to attach BPF program %s to kretprobe %s" %
+                raise Exception("Failed to attach BPF program %s to kretprobe %s"
+                                ", it's not traceable (either non-existing, inlined, or marked as \"notrace\")" %
                                 (fn_name, '/'.join(probes)))
             return
 
@@ -862,7 +873,8 @@ class BPF(object):
         ev_name = b"r_" + event.replace(b"+", b"_").replace(b".", b"_")
         fd = lib.bpf_attach_kprobe(fn.fd, 1, ev_name, event, 0, maxactive)
         if fd < 0:
-            raise Exception("Failed to attach BPF program %s to kretprobe %s" %
+            raise Exception("Failed to attach BPF program %s to kretprobe %s"
+                            ", it's not traceable (either non-existing, inlined, or marked as \"notrace\")" %
                             (fn_name, event))
         self._add_kprobe_fd(ev_name, fn_name, fd)
         return self
@@ -896,7 +908,6 @@ class BPF(object):
         else:
             self.detach_kprobe_event(ev_name)
 
-
     def detach_kretprobe(self, event, fn_name=None):
         event = _assert_is_bytes(event)
         ev_name = b"r_" + event.replace(b"+", b"_").replace(b".", b"_")
@@ -939,8 +950,6 @@ class BPF(object):
             raise Exception("Failed to detach BPF from device %s: %s"
                             % (dev, errstr))
 
-
-
     @classmethod
     def _check_path_symbol(cls, module, symname, addr, pid, sym_off=0):
         module = _assert_is_bytes(module)
@@ -953,7 +962,8 @@ class BPF(object):
             ct.cast(None, ct.POINTER(bcc_symbol_option)),
             ct.byref(sym),
         ) < 0:
-            raise Exception("could not determine address of symbol %s" % symname)
+            raise Exception("could not determine address of symbol %s in %s"
+                            % (symname.decode(), module.decode()))
         new_addr = sym.offset + sym_off
         module_path = ct.cast(sym.module, ct.c_char_p).value
         lib.bcc_procutils_free(sym.module)
@@ -1183,6 +1193,12 @@ class BPF(object):
                     return True
             return False
 
+    @staticmethod
+    def kernel_struct_has_field(struct_name, field_name):
+        struct_name = _assert_is_bytes(struct_name)
+        field_name = _assert_is_bytes(field_name)
+        return lib.kernel_struct_has_field(struct_name, field_name)
+
     def detach_tracepoint(self, tp=b""):
         """detach_tracepoint(tp="")
 
@@ -1657,6 +1673,18 @@ class BPF(object):
             readers[i] = v
         lib.perf_reader_poll(len(readers), readers, timeout)
 
+    def perf_buffer_consume(self):
+        """perf_buffer_consume(self)
+
+        Consume all open perf buffers, regardless of whether or not
+        they currently contain events data. Necessary to catch 'remainder'
+        events when wakeup_events > 1 is set in open_perf_buffer
+        """
+        readers = (ct.c_void_p * len(self.perf_buffers))()
+        for i, v in enumerate(self.perf_buffers.values()):
+            readers[i] = v
+        lib.perf_reader_consume(len(readers), readers)
+
     def kprobe_poll(self, timeout = -1):
         """kprobe_poll(self)
 
@@ -1713,6 +1741,20 @@ class BPF(object):
     def donothing(self):
         """the do nothing exit handler"""
 
+
+    def close(self):
+        """close(self)
+
+        Closes all associated files descriptors. Attached BPF programs are not
+        detached.
+        """
+        for name, fn in list(self.funcs.items()):
+            os.close(fn.fd)
+            del self.funcs[name]
+        if self.module:
+            lib.bpf_module_destroy(self.module)
+            self.module = None
+
     def cleanup(self):
         # Clean up opened probes
         for k, v in list(self.kprobe_fds.items()):
@@ -1740,12 +1782,8 @@ class BPF(object):
         if self.tracefile:
             self.tracefile.close()
             self.tracefile = None
-        for name, fn in list(self.funcs.items()):
-            os.close(fn.fd)
-            del self.funcs[name]
-        if self.module:
-            lib.bpf_module_destroy(self.module)
-            self.module = None
+
+        self.close()
 
         # Clean up ringbuf
         if self._ringbuf_manager:
